import json
from abc import ABCMeta, abstractmethod
from collections import namedtuple

from six import add_metaclass

from endpoints.api import format_date


class Tag(
  namedtuple('Tag', [
    'name', 'image', 'reversion', 'lifetime_start_ts', 'lifetime_end_ts', 'manifest_list',
    'docker_image_id'
  ])):
  """
  Tag represents a name to an image.
  :type name: string
  :type image: Image
  :type reversion: boolean
  :type lifetime_start_ts: int
  :type lifetime_end_ts: int
  :type manifest_list: [manifest_digest]
  :type docker_image_id: string
  """

  def to_dict(self):
    tag_info = {
      'name': self.name,
      'docker_image_id': self.docker_image_id,
      'reversion': self.reversion,
    }

    if self.lifetime_start_ts > 0:
      tag_info['start_ts'] = self.lifetime_start_ts

    if self.lifetime_end_ts > 0:
      tag_info['end_ts'] = self.lifetime_end_ts

    if self.manifest_list:
      tag_info['manifest_digest'] = self.manifest_list

    return tag_info


class RepositoryTagHistory(namedtuple('RepositoryTagHistory', ['tags', 'more'])):
  """
  Tag represents a name to an image.
  :type tags: [Tag]
  :type more: boolean
  """


class Repository(namedtuple('Repository', ['namespace_name', 'repository_name'])):
  """
  Repository a single quay repository
  :type namespace_name: string
  :type repository_name: string
  """


class Image(
  namedtuple('Image', [
    'docker_image_id', 'created', 'comment', 'command', 'storage_image_size',
    'storage_uploading', 'ancestor_id_list', 'ancestor_id'
  ])):
  """
  Image
  :type docker_image_id: string
  :type created: datetime
  :type comment: string
  :type command: string
  :type storage_image_size: int
  :type storage_uploading: boolean
  :type ancestor_id_list: [int]
  :type ancestor_id: int
  """

  def to_dict(self, image_map, include_ancestors=True):
    command = self.command

    def docker_id(aid):
      if aid not in image_map:
        return ''

      return image_map[aid].docker_image_id

    image_data = {
      'id': self.docker_image_id,
      'created': format_date(self.created),
      'comment': self.comment,
      'command': json.loads(command) if command else None,
      'size': self.storage_image_size,
      'uploading': self.storage_uploading,
      'sort_index': len(self.ancestor_id_list),
    }

    if include_ancestors:
      # Calculate the ancestors string, with the DBID's replaced with the docker IDs.
      ancestors = [docker_id(a) for a in self.ancestor_id_list]
      image_data['ancestors'] = '/{0}/'.format('/'.join(ancestors))

    return image_data


@add_metaclass(ABCMeta)
class TagDataInterface(object):
  """
  Interface that represents all data store interactions required by a Tag.
  """

  @abstractmethod
  def list_repository_tag_history(self, namespace_name, repository_name, page=1, size=100,
                                  specific_tag=None):
    """
    Returns a RepositoryTagHistory with a list of historic tags and whether there are more tags then returned.
    """

  @abstractmethod
  def get_repo(self, namespace_name, repository_name):
    """
    Returns a repository associated with the given namespace and repository name.
    """

  @abstractmethod
  def get_repo_tag_image(self, repository, tag_name):
    """
    Returns an image associated with the repository and tag_name
    """

  @abstractmethod
  def create_or_update_tag(self, namespace_name, repository_name, tag_name, docker_image_id):
    """
    Returns the repository tag if it is created.
    """

  @abstractmethod
  def delete_tag(self, namespace_name, repository_name, tag_name):
    """
    Returns the tag for the given namespace and repository if it was created
    """

  @abstractmethod
  def get_parent_images(self, namespace, repository, tag_name):
    """
    Returns a list of the parent images for the namespace, repository and tag specified.
    """

  @abstractmethod
  def list_repository_tags(self, namespace_name, repository_name):
    """
    Returns a list of all tags associated with namespace_nam and repository_name
    """

  @abstractmethod
  def get_repository(self, namespace_name, repository_name):
    """
    Returns the repository associated with the namespace_name and repository_name
    """

  @abstractmethod
  def get_repository_image(self, namespace_name, repository_name, docker_image_id):
    """
    Returns the repository image associated with the namespace_name, repository_name, and docker
    image ID.
    """

  @abstractmethod
  def restore_tag_to_manifest(self, repository_name, tag_name, manifest_digest):
    """
    Returns the existing repo tag image if it exists or else returns None.
    Side effects include adding the tag with associated name to the manifest_digest in the named repo.
    """

  @abstractmethod
  def restore_tag_to_image(self, repository_name, tag_name, image_id):
    """
    Returns the existing repo tag image if it exists or else returns None
    Side effects include adding the tag with associated name to the image with the associated id in the named repo.
    """

  @abstractmethod
  def change_repository_tag_expiration(self, namespace_name, repository_name, tag_name,
                                       expiration_date):
    """ Sets the expiration date of the tag under the matching repository to that given. If the
        expiration date is None, then the tag will not expire. Returns a tuple of the previous
        expiration timestamp in seconds (if any), and whether the operation succeeded.
    """
