from data import model
from data.model import DataModelException, InvalidImageException
from endpoints.api.tag_models_interface import TagDataInterface, Tag, RepositoryTagHistory, Repository, Image


class PreOCIModel(TagDataInterface):
  """
  PreOCIModel implements the data model for the Tags using a database schema
  before it was changed to support the OCI specification.
  """

  def list_repository_tag_history(self, namespace_name, repository_name, page=1, size=100,
                                  specific_tag=None):
    repository = model.repository.get_repository(namespace_name, repository_name)
    if repository is None:
      return None

    tags, manifest_map, more = model.tag.list_repository_tag_history(repository, page, size,
                                                                     specific_tag)
    repository_tag_history = []
    for tag in tags:
      manifest_list = None
      if tag.id in manifest_map:
        manifest_list = manifest_map[tag.id]

      repository_tag_history.append(convert_tag(tag, manifest_list))

    return RepositoryTagHistory(tags=repository_tag_history, more=more)

  def get_repo(self, namespace_name, repository_name):
    repo = model.repository.get_repository(namespace_name, repository_name)
    if repo is None:
      return None

    return Repository(repo.namespace_user, repo.name)

  def get_repo_tag_image(self, repository, tag_name):
    repo = model.repository.get_repository(str(repository.namespace_name), str(repository.repository_name))
    if repo is None:
      return None

    try:
      image = model.tag.get_repo_tag_image(repo, tag_name)
    except DataModelException:
      return None

    return convert_image(image)

  def create_or_update_tag(self, namespace_name, repository_name, tag_name, docker_image_id):
    return model.tag.create_or_update_tag(namespace_name, repository_name, tag_name,
                                          docker_image_id)

  def delete_tag(self, namespace_name, repository_name, tag_name):
    return model.tag.delete_tag(namespace_name, repository_name, tag_name)

  def get_parent_images(self, namespace_name, repository_name, docker_image_id):
    try:
      image = model.image.get_image_by_id(namespace_name, repository_name, docker_image_id)
    except InvalidImageException:
      return []

    parent_tags = model.image.get_parent_images(namespace_name, repository_name, image)
    return_tags = []
    for image in parent_tags:
      return_tags.append(convert_image(image))

    return return_tags

  def list_repository_tags(self, namespace_name, repository_name):
    tags = model.tag.list_repository_tags(namespace_name, repository_name)
    new_tags = []
    for tag in tags:
      new_tags.append(convert_tag(tag))
    return new_tags

  def get_repository_image(self, namespace_name, repository_name, docker_image_id):
    image = model.image.get_repo_image(namespace_name, repository_name, docker_image_id)
    if image is None:
      return None

    return convert_image(image)

  def get_repository(self, namespace_name, repository_name):
    repo = model.repository.get_repository(namespace_name, repository_name)
    if repo is None:
      return None

    return Repository(namespace_name=namespace_name, repository_name=repository_name)

  def restore_tag_to_manifest(self, repository, tag_name, manifest_digest):
    repo = model.repository.get_repository(repository.namespace_name, repository.repository_name)
    if repo is None:
      return None

    image = model.tag.restore_tag_to_manifest(repo, tag_name, manifest_digest)
    if image is None:
      return None

    return convert_image(image)

  def restore_tag_to_image(self, repository, tag_name, image_id):
    repo = model.repository.get_repository(repository.namespace_name, repository.repository_name)
    if repo is None:
      return None

    image = model.tag.restore_tag_to_image(repo, tag_name, image_id)
    if image is None:
      return None

    return convert_image(image)

  def change_repository_tag_expiration(self, namespace_name, repository_name, tag_name,
                                       expiration_date):
    return model.tag.change_repository_tag_expiration(namespace_name, repository_name, tag_name,
                                                      expiration_date)


def convert_image(database_image):
  return Image(docker_image_id=database_image.docker_image_id, created=database_image.created,
               comment=database_image.comment, command=database_image.command,
               storage_image_size=database_image.storage.image_size,
               storage_uploading=database_image.storage.uploading,
               ancestor_id_list=database_image.ancestor_id_list(),
               ancestor_id=database_image.id)


def convert_tag(tag, manifest_list=None):
  return Tag(name=tag.name, image=convert_image(tag.image), reversion=tag.reversion,
             lifetime_start_ts=tag.lifetime_start_ts, lifetime_end_ts=tag.lifetime_end_ts,
             manifest_list=manifest_list, docker_image_id=tag.image.docker_image_id)


pre_oci_model = PreOCIModel()
