import json

import pytest

from mock import patch, Mock, MagicMock, call

from data.model import DataModelException
from endpoints.api.tag_models_interface import RepositoryTagHistory, Tag
from endpoints.api.test.shared import conduct_api_call
from endpoints.test.shared import client_with_identity
from endpoints.api.tag import RepositoryTag, RestoreTag, ListRepositoryTags, RepositoryTagImages

from features import FeatureNameValue

from test.fixtures import *


@pytest.fixture()
def get_repo_image():
  def mock_callable(namespace, repository, image_id):
    mock = Mock(namespace_user='devtable')
    mock.name = 'simple'
    img = Mock(repository=mock, docker_image_id=12) if image_id == 'image1' else None
    return img

  with patch('endpoints.api.tag_models_pre_oci.model.image.get_repo_image',
             side_effect=mock_callable) as mk:
    yield mk


@pytest.fixture()
def get_repository():
  with patch('endpoints.api.tag_models_pre_oci.model.image.get_repo_image',
             return_value='mock_repo') as mk:
    yield mk


@pytest.fixture()
def get_repo_tag_image():
  def mock_get_repo_tag_image(repository, tag):
    storage_mock = Mock(image_size=1234, uploading='uploading')

    def fake_ancestor_id_list():
      return []

    if tag == 'existing-tag':
      return Mock(docker_image_id='mock_docker_image_id', created=12345, comment='comment',
                  command='command', storage=storage_mock, ancestors=[],
                  ancestor_id_list=fake_ancestor_id_list)
    else:
      raise DataModelException('Unable to find image for tag.')

  with patch('endpoints.api.tag_models_pre_oci.model.tag.get_repo_tag_image',
             side_effect=mock_get_repo_tag_image):
    yield


@pytest.fixture()
def restore_tag_to_manifest():
  def mock_restore_tag_to_manifest(repository, tag, manifest_digest):
    tag_img = Mock(docker_image_id='mock_docker_image_id') if tag == 'existing-tag' else None
    return tag_img

  with patch('endpoints.api.tag_models_pre_oci.model.tag.restore_tag_to_manifest',
             side_effect=mock_restore_tag_to_manifest):
    yield


@pytest.fixture()
def restore_tag_to_image():
  def mock_restore_tag_to_image(repository, tag, image_id):
    tag_img = Mock(docker_image_id='mock_docker_image_id') if tag == 'existing-tag' else None
    return tag_img

  with patch('endpoints.api.tag_models_pre_oci.model.tag.restore_tag_to_image',
             side_effect=mock_restore_tag_to_image):
    yield


@pytest.fixture()
def create_or_update_tag():
  with patch('endpoints.api.tag_models_pre_oci.model.tag.create_or_update_tag') as mk:
    yield mk


@pytest.fixture()
def generate_manifest():
  def mock_callable(namespace, repository, tag):
    if tag == 'generatemanifestfail':
      raise Exception('test_failure')

  with patch('endpoints.api.tag._generate_and_store_manifest', side_effect=mock_callable) as mk:
    yield mk


@pytest.fixture()
def authd_client(client):
  with client_with_identity('devtable', client) as cl:
    yield cl


@pytest.fixture()
def list_repository_tag_history():
  def list_repository_tag_history(namespace_name, repository_name, page, size, specific_tag):
    return RepositoryTagHistory(tags=[
      Tag(name='First Tag', image='image', reversion=False, lifetime_start_ts=0, lifetime_end_ts=0,
          manifest_list=[], docker_image_id='first docker image id'),
      Tag(name='Second Tag', image='second image', reversion=True, lifetime_start_ts=10,
          lifetime_end_ts=100, manifest_list=[], docker_image_id='second docker image id')
    ], more=False)

  with patch('endpoints.api.tag.model.list_repository_tag_history',
             side_effect=list_repository_tag_history):
    yield


@pytest.fixture()
def find_no_repo_tag_history():
  def list_repository_tag_history(namespace_name, repository_name, page, size, specific_tag):
    return None

  with patch('endpoints.api.tag.model.list_repository_tag_history',
             side_effect=list_repository_tag_history):
    yield


@pytest.mark.parametrize('expiration_time, expected_status', [
  (None, 201),
  ('aksdjhasd', 400),
])
def test_change_tag_expiration_default(expiration_time, expected_status, client, app):
  with client_with_identity('devtable', client) as cl:
    params = {
      'repository': 'devtable/simple',
      'tag': 'latest',
    }

    request_body = {
      'expiration': expiration_time,
    }

    conduct_api_call(cl, RepositoryTag, 'put', params, request_body, expected_status)


def test_change_tag_expiration(client, app):
  with client_with_identity('devtable', client) as cl:
    params = {
      'repository': 'devtable/simple',
      'tag': 'latest',
    }

    tag = model.tag.get_active_tag('devtable', 'simple', 'latest')
    updated_expiration = tag.lifetime_start_ts + 60*60*24

    request_body = {
      'expiration': updated_expiration,
    }

    conduct_api_call(cl, RepositoryTag, 'put', params, request_body, 201)
    tag = model.tag.get_active_tag('devtable', 'simple', 'latest')
    assert tag.lifetime_end_ts == updated_expiration


@pytest.mark.parametrize('test_image,test_tag,expected_status', [
  ('image1', '-INVALID-TAG-NAME', 400),
  ('image1', '.INVALID-TAG-NAME', 400),
  ('image1',
   'INVALID-TAG_NAME-BECAUSE-THIS-IS-WAY-WAY-TOO-LOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOONG',
   400),
  ('nonexistantimage', 'newtag', 404),
  ('image1', 'generatemanifestfail', None),
  ('image1', 'existing-tag', 201),
  ('image1', 'newtag', 201),
])
def test_move_tag(test_image, test_tag, expected_status, get_repo_image, get_repo_tag_image,
                  create_or_update_tag, generate_manifest, authd_client):
  params = {'repository': 'devtable/simple', 'tag': test_tag}
  request_body = {'image': test_image}
  if expected_status is None:
    with pytest.raises(Exception):
      conduct_api_call(authd_client, RepositoryTag, 'put', params, request_body, expected_status)
  else:
    conduct_api_call(authd_client, RepositoryTag, 'put', params, request_body, expected_status)


@pytest.mark.parametrize(
  'namespace, repository, specific_tag, page, limit, expected_response_code, expected', [
    ('devtable', 'simple', None, 1, 10, 200, {
      'has_additional': False
    }),
    ('devtable', 'simple', None, 1, 10, 200, {
      'page': 1
    }),
    ('devtable', 'simple', None, 1, 10, 200, {
      'tags': [{
        'docker_image_id': 'first docker image id',
        'name': 'First Tag',
        'reversion': False
      }, {
        'docker_image_id': 'second docker image id',
        'end_ts': 100,
        'name': 'Second Tag',
        'reversion': True,
        'start_ts': 10
      }]
    }),
  ])
def test_list_repository_tags_view_is_correct(namespace, repository, specific_tag, page, limit,
                                              list_repository_tag_history, expected_response_code,
                                              expected, authd_client):
  params = {
    'repository': namespace + '/' + repository,
    'specificTag': specific_tag,
    'page': page,
    'limit': limit
  }
  response = conduct_api_call(authd_client, ListRepositoryTags, 'get', params,
                              expected_code=expected_response_code)
  compare_list_history_tags_response(expected, response.json)


def compare_list_history_tags_response(expected, actual):
  if 'has_additional' in expected:
    assert expected['has_additional'] == actual['has_additional']

  if 'page' in expected:
    assert expected['page'] == actual['page']

  if 'tags' in expected:
    assert expected['tags'] == actual['tags']


def test_no_repo_tag_history(find_no_repo_tag_history, authd_client):
  params = {'repository': 'devtable/simple', 'specificTag': None, 'page': 1, 'limit': 10}
  conduct_api_call(authd_client, ListRepositoryTags, 'get', params, expected_code=404)


@pytest.mark.parametrize(
  'specific_tag, page, limit, expected_specific_tag, expected_page, expected_limit', [
    (None, None, None, None, 1, 50),
    ('specific_tag', 12, 13, 'specific_tag', 12, 13),
    ('specific_tag', -1, 101, 'specific_tag', 1, 100),
    ('specific_tag', 0, 0, 'specific_tag', 1, 1),
  ])
def test_repo_tag_history_param_parse(specific_tag, page, limit, expected_specific_tag,
                                      expected_page, expected_limit, authd_client):
  mock = MagicMock()
  mock.return_value = RepositoryTagHistory(tags=[], more=False)

  with patch('endpoints.api.tag.model.list_repository_tag_history', side_effect=mock):
    params = {
      'repository': 'devtable/simple',
      'specificTag': specific_tag,
      'page': page,
      'limit': limit
    }
    conduct_api_call(authd_client, ListRepositoryTags, 'get', params)

  assert mock.call_args == call(namespace_name='devtable', repository_name='simple',
                                page=expected_page, size=expected_limit,
                                specific_tag=expected_specific_tag)


@pytest.mark.parametrize('test_manifest,test_tag,manifest_generated,expected_status', [
  (None, 'newtag', True, 200),
  (None, 'generatemanifestfail', True, None),
  ('manifest1', 'newtag', False, 200),
])
def test_restore_tag(test_manifest, test_tag, manifest_generated, expected_status, get_repository,
                     restore_tag_to_manifest, restore_tag_to_image, generate_manifest,
                     authd_client):
  params = {'repository': 'devtable/simple', 'tag': test_tag}
  request_body = {'image': 'image1'}
  if test_manifest is not None:
    request_body['manifest_digest'] = test_manifest
  if expected_status is None:
    with pytest.raises(Exception):
      conduct_api_call(authd_client, RestoreTag, 'post', params, request_body, expected_status)
  else:
    conduct_api_call(authd_client, RestoreTag, 'post', params, request_body, expected_status)

  if manifest_generated:
    generate_manifest.assert_called_with('devtable', 'simple', test_tag)

@pytest.mark.parametrize('repository, tag, owned, expect_images', [
  ('devtable/simple', 'prod', False, True),
  ('devtable/simple', 'prod', True, False),
  ('devtable/simple', 'latest', False, True),
  ('devtable/simple', 'latest', True, False),

  ('devtable/complex', 'prod', False, True),
  ('devtable/complex', 'prod', True, True),
])
def test_list_tag_images(repository, tag, owned, expect_images, authd_client):
  params = {'repository': repository, 'tag': tag, 'owned': owned}
  result = conduct_api_call(authd_client, RepositoryTagImages, 'get', params, None, 200).json
  assert bool(result['images']) == expect_images
