import pytest

from data.model import DataModelException, InvalidImageException
from endpoints.api.tag_models_interface import RepositoryTagHistory, Tag, Repository
from mock import Mock, call

from data import model
from endpoints.api.tag_models_pre_oci import pre_oci_model
from util.morecollections import AttrDict

EMPTY_REPOSITORY = 'empty_repository'
EMPTY_NAMESPACE = 'empty_namespace'
BAD_REPOSITORY_NAME = 'bad_repository_name'
BAD_NAMESPACE_NAME = 'bad_namespace_name'


@pytest.fixture
def get_monkeypatch(monkeypatch):
  return monkeypatch


def mock_out_get_repository(monkeypatch, namespace_name, repository_name):
  def return_none(namespace_name, repository_name):
    return None

  def return_repository(namespace_name, repository_name):
    return 'repository'

  if namespace_name == BAD_NAMESPACE_NAME or repository_name == BAD_REPOSITORY_NAME:
    return_function = return_none
  else:
    return_function = return_repository

  monkeypatch.setattr(model.repository, 'get_repository', return_function)


def get_repo_mock(monkeypatch, return_value):
  def return_return_value(namespace_name, repository_name):
    return return_value

  monkeypatch.setattr(model.repository, 'get_repository', return_return_value)


def test_get_repo_not_exists(get_monkeypatch):
  namespace_name = 'namespace_name'
  repository_name = 'repository_name'
  get_repo_mock(get_monkeypatch, None)
  repo = pre_oci_model.get_repo(namespace_name, repository_name)

  assert repo is None


def test_get_repo_exists(get_monkeypatch):
  namespace_name = 'namespace_name'
  repository_name = 'repository_name'
  mock = Mock()
  mock.namespace_user = namespace_name
  mock.name = repository_name
  mock.repository = mock
  get_repo_mock(get_monkeypatch, mock)

  repo = pre_oci_model.get_repo(namespace_name, repository_name)

  assert repo is not None
  assert repo.repository_name == repository_name
  assert repo.namespace_name == namespace_name


def get_repository_mock(monkeypatch, return_value):
  def return_return_value(namespace_name, repository_name, kind_filter=None):
    return return_value

  monkeypatch.setattr(model.repository, 'get_repository', return_return_value)


def get_repo_tag_image_mock(monkeypatch, return_value):
  def return_return_value(repo, tag_name, include_storage=False):
    return return_value

  monkeypatch.setattr(model.tag, 'get_repo_tag_image', return_return_value)


def test_get_repo_tag_image_with_repo_and_repo_tag(get_monkeypatch):
  mock_storage = Mock()
  mock_image = Mock()
  mock_image.docker_image_id = 'some docker image id'
  mock_image.created = 1235
  mock_image.comment = 'some comment'
  mock_image.command = 'some command'
  mock_image.storage = mock_storage
  mock_image.ancestors = []

  get_repository_mock(get_monkeypatch, mock_image)
  get_repo_tag_image_mock(get_monkeypatch, mock_image)

  image = pre_oci_model.get_repo_tag_image(
    Repository('namespace_name', 'repository_name'), 'tag_name')

  assert image is not None
  assert image.docker_image_id == 'some docker image id'


def test_get_repo_tag_image_without_repo(get_monkeypatch):
  get_repository_mock(get_monkeypatch, None)

  image = pre_oci_model.get_repo_tag_image(
    Repository('namespace_name', 'repository_name'), 'tag_name')

  assert image is None


def test_get_repo_tag_image_without_repo_tag_image(get_monkeypatch):
  mock = Mock()
  mock.docker_image_id = 'some docker image id'
  get_repository_mock(get_monkeypatch, mock)

  def raise_exception(repo, tag_name, include_storage=False):
    raise DataModelException()

  get_monkeypatch.setattr(model.tag, 'get_repo_tag_image', raise_exception)

  image = pre_oci_model.get_repo_tag_image(
    Repository('namespace_name', 'repository_name'), 'tag_name')

  assert image is None


def test_create_or_update_tag(get_monkeypatch):
  mock = Mock()
  get_monkeypatch.setattr(model.tag, 'create_or_update_tag', mock)

  pre_oci_model.create_or_update_tag('namespace_name', 'repository_name', 'tag_name',
                                     'docker_image_id')

  assert mock.call_count == 1
  assert mock.call_args == call('namespace_name', 'repository_name', 'tag_name', 'docker_image_id')


def test_delete_tag(get_monkeypatch):
  mock = Mock()
  get_monkeypatch.setattr(model.tag, 'delete_tag', mock)

  pre_oci_model.delete_tag('namespace_name', 'repository_name', 'tag_name')

  assert mock.call_count == 1
  assert mock.call_args == call('namespace_name', 'repository_name', 'tag_name')


def test_get_parent_images_with_exception(get_monkeypatch):
  mock = Mock(side_effect=InvalidImageException)

  get_monkeypatch.setattr(model.image, 'get_image_by_id', mock)

  images = pre_oci_model.get_parent_images('namespace_name', 'repository_name', 'tag_name')
  assert images == []


def test_get_parent_images_empty_parent_images(get_monkeypatch):
  get_image_by_id_mock = Mock()
  get_monkeypatch.setattr(model.image, 'get_image_by_id', get_image_by_id_mock)

  get_parent_images_mock = Mock(return_value=[])
  get_monkeypatch.setattr(model.image, 'get_parent_images', get_parent_images_mock)

  images = pre_oci_model.get_parent_images('namespace_name', 'repository_name', 'tag_name')
  assert images == []


def test_list_repository_tags(get_monkeypatch):
  mock = Mock(return_value=[])

  get_monkeypatch.setattr(model.tag, 'list_repository_tags', mock)

  pre_oci_model.list_repository_tags('namespace_name', 'repository_name')

  mock.assert_called_once_with('namespace_name', 'repository_name')


def test_get_repository(get_monkeypatch):
  mock = Mock()

  get_monkeypatch.setattr(model.repository, 'get_repository', mock)

  pre_oci_model.get_repository('namespace_name', 'repository_name')

  mock.assert_called_once_with('namespace_name', 'repository_name')


def test_tag_to_manifest(get_monkeypatch):
  repo_mock = Mock()
  restore_tag_mock = Mock(return_value=None)
  get_repository_mock = Mock(return_value=repo_mock)

  get_monkeypatch.setattr(model.tag, 'restore_tag_to_manifest', restore_tag_mock)
  get_monkeypatch.setattr(model.repository, 'get_repository', get_repository_mock)

  pre_oci_model.restore_tag_to_manifest(
    Repository('namespace', 'repository'), 'tag_name', 'manifest_digest')

  get_repository_mock.assert_called_once_with('namespace', 'repository')
  restore_tag_mock.assert_called_once_with(repo_mock, 'tag_name', 'manifest_digest')


def test__tag_to_image(get_monkeypatch):
  repo_mock = Mock()
  restore_tag_mock = Mock(return_value=None)
  get_repository_mock = Mock(return_value=repo_mock)

  get_monkeypatch.setattr(model.tag, 'restore_tag_to_image', restore_tag_mock)
  get_monkeypatch.setattr(model.repository, 'get_repository', get_repository_mock)

  pre_oci_model.restore_tag_to_image(Repository('namespace', 'repository'), 'tag_name', 'image_id')

  get_repository_mock.assert_called_once_with('namespace', 'repository')
  restore_tag_mock.assert_called_once_with(repo_mock, 'tag_name', 'image_id')
